/*

linux/fs/dmsdos/dmsdos_compr.c

DMSDOS filesystem: compression routines

******************************************************************************
DMSDOS (Doublespace/Drivespace compressed MSDOS filesystem) for Linux
written 1995,1996 by Frank Gockel

    (C) Copyright 1995-1997 by Frank Gockel
    (C) Copyright 1996,1997 by Pavel Pisa

Some code of the dmsdos filesystem has been copied from the msdos filesystem
so there are the following additional copyrights:

    (C) Copyright 1992,1993 by Werner Almesberger (msdos filesystem)
    (C) Copyright 1994,1995 by Jacques Gelinas (mmap code)
    (C) Copyright 1992-1995 by Linus Torvalds

The DMSDOS filesystem was inspired by the THS filesystem (a simple doublespace
DS-0-2 compressed read-only filesystem) written 1994 by Thomas Scheuermann.

The DMSDOS filesystem is distributed under the Gnu General Public Licence.
See file COPYING for details.
******************************************************************************

*/

#include <linux/errno.h>
#include <linux/kernel.h>
#include <linux/malloc.h>
#include <linux/fs.h>
#include <linux/msdos_fs.h>
#include "dmsdos.h"

#ifdef __KERNEL__
extern unsigned long dmsdos_speedup;
#else
extern int cfaktor;
void panic(char*);
#include <stdlib.h>
#define MALLOC malloc
#define FREE   free
#endif /*__KERNEL__*/

/* maybe the values have to be adapted for optimal compression/speed 
   relationship */
int c_maxdiff[12]={ 63, 63, 63, 63,319,319,319,319,4414,4414,4414,4414};
int c_maxlen[12] ={ 64,128,256,512, 64,128,256,512,  64, 128, 256, 512};

#define INLINE  extern inline

#include <asm/byteorder.h>
#if !defined(cpu_to_le16) && ((CPU==386)||(CPU==486)||(CPU==586)||(CPU==686))
 /* for old kernel versions - works only on i386 */
 #define cpu_to_le16(v) (v)
#endif

/* for reading and writting from/to bitstream */
typedef
 struct {
   __u32 buf;	/* bit buffer */
     int pb;	/* already written bits to buf */
   __u16 *pd;	/* data write pointer */
   __u16 *pe;	/* after end of data */
 } bits_t;

/* initializes writting to bitstream */
INLINE void dblb_wri(bits_t *pbits,void *pin,unsigned lin)
{
  pbits->buf=0;
  pbits->pb=0;
  pbits->pd=(__u16*)pin;
  pbits->pe=pbits->pd+((lin+1)>>1);
}

/* writes n<=16 bits to bitstream *pbits */
INLINE void dblb_wrn(bits_t *pbits,int cod, int n)
{
  pbits->buf|=cod<<pbits->pb;
  if((pbits->pb+=n)>=16)
  {
    if(pbits->pd<pbits->pe)
      *(pbits->pd++)=(__u16)pbits->buf;
    else if(pbits->pd==pbits->pe) pbits->pd++; /* output overflow */
    pbits->buf>>=16;
    pbits->pb-=16;
  }
}

void write_byte(bits_t *pbits,int byte,int method)
{ if(method!=JM_0_0&&method!=JM_0_1)
  { if(byte<128)dblb_wrn(pbits,2,2);
    else dblb_wrn(pbits,1,2);
  }
  else /* JM_0_0 */
  { if(byte<128)dblb_wrn(pbits,0,1);
    else dblb_wrn(pbits,3,2);
  }
  dblb_wrn(pbits,byte&0x7F,7);
}

void write_temp(bits_t *pbits,
                unsigned char*tempstr,int len,int diff,int method)
{ if(len==1)
  { write_byte(pbits,tempstr[0],method);
    return;
  }
  if(len==2&&(method==JM_0_0||method==JM_0_1))
  { write_byte(pbits,tempstr[0],method);
    write_byte(pbits,tempstr[1],method);
    return;
  }
  if(method!=JM_0_0&&method!=JM_0_1)
  { ++len; /* corrects different counting scheme */
    if(diff<64)dblb_wrn(pbits,0,2);
    else dblb_wrn(pbits,3,2);
  }
  else /* JM_0_0 */
  { dblb_wrn(pbits,1,2);
    dblb_wrn(pbits,(diff<64)?0:1,1);
  }
  if(diff<64)dblb_wrn(pbits,diff,6);
  else 
  { if(diff<320)
    { dblb_wrn(pbits,0,1);
      dblb_wrn(pbits,diff-64,8);
    }
    else
    { dblb_wrn(pbits,1,1);
      dblb_wrn(pbits,diff-320,12);
    }
  }
  /* okay, now encode len */
  if(len==3)
  { dblb_wrn(pbits,1,1);
    return;
  }
  if(len<6)
  { dblb_wrn(pbits,1<<1,2);
    dblb_wrn(pbits,len-4,1);
    return;
  }
  if(len<10)
  { dblb_wrn(pbits,1<<2,3);
    dblb_wrn(pbits,len-6,2);
    return;
  }
  if(len<18)
  { dblb_wrn(pbits,1<<3,4);
    dblb_wrn(pbits,len-10,3);
    return;
  }
  if(len<34)
  { dblb_wrn(pbits,1<<4,5);
    dblb_wrn(pbits,len-18,4);
    return;
  }
  if(len<66)
  { dblb_wrn(pbits,1<<5,6);
    dblb_wrn(pbits,len-34,5);
    return;
  }
  if(len<130)
  { dblb_wrn(pbits,1<<6,7);
    dblb_wrn(pbits,len-66,6);
    return;
  }
  if(len<258)
  { dblb_wrn(pbits,1<<7,8);
    dblb_wrn(pbits,len-130,7);
    return;
  }
  dblb_wrn(pbits,1<<8,9);
  dblb_wrn(pbits,len-258,8);
}

void write_marker(bits_t *pbits,int method)
{ if(method==JM_0_0||method==JM_0_1)dblb_wrn(pbits,13,4);
  else /* DS_0_x */ dblb_wrn(pbits,7,3);
  dblb_wrn(pbits,0xFFF,12); 
}

typedef __u16 hash_t;

/* hashing function is only 2 char because of DS min rep */
INLINE unsigned dbl_hash(__u8 *p)
{
 return (((__u16)p[0]<<4)^((__u16)p[1]<<0))&0x3FF;
};

/* adds new hash and returns previous occurence */
INLINE hash_t dbl_newhash(__u8*clusterd ,int pos,
                     hash_t* hash_tab,hash_t* hash_hist,unsigned hash_mask)
{
 hash_t* hash_ptr;
 hash_t  hash_cur;
 hash_ptr=hash_tab+dbl_hash(&(clusterd[pos]));
 hash_cur=*hash_ptr;		/* find previous occurence of hash */
 *hash_ptr=pos;			/* store new occurence */
 *(hash_hist+(pos&hash_mask))=hash_cur; 
                                /* store previous in history table */
 return(hash_cur);
};

/* compresses a doublespace/drivespace cluster 
   gets uncompressed size (number of used sectors)
   returns compressed size (in number of used sectors) or -1 if failed
*/
int dbl_compress(__u8* clusterk, __u8* clusterd, int size,
             int method,int cf)
{ 
  bits_t bits;
  int i;
  int pos;
  int mark_pos; /* position of next marker */
  hash_t *hash_tab; 
        /* [0x400] pointers to last occurrence of same hash, index hash */
  hash_t *hash_hist;
        /* [0x800] previous occurences of hash, index actual pointer&hash_mask */ 
  unsigned hash_mask=0x7FF;/* mask for index into hash_hist */
  int hash_cur;
  int max_hash=0; /* GCC likes this */
  int match;
  int max_match;
  int try_cn;
  int try_count=10; /* tunable parameter */

  switch(method)
  { case DS_0_0:
    case DS_0_1:
    case DS_0_2: /* handled together with JM_0_0 because similar */
    case JM_0_0:
    case JM_0_1:

      /* maximal compression */
      if(cf>8)hash_mask=0xFFF;

      /* Input size in bytes */
      size=size*SECTOR_SIZE;

      /* initialize bitstream */
      dblb_wri(&bits,clusterk,size-SECTOR_SIZE);
            
      /* put magic number */
      dblb_wrn(&bits,method,32);
      
      hash_tab=MALLOC(0x400*sizeof(hash_t));
      if(hash_tab==NULL) return -1;
      hash_hist=MALLOC((hash_mask+1)*sizeof(hash_t));
      if(hash_hist==NULL) {FREE(hash_tab);return -1;}
      
      for(i=0;i<0x400;i++) hash_tab[i]=0xFFFF;
      for(i=0;i<=hash_mask;i++) hash_hist[i]=0xFFFF;
      
      pos=mark_pos=0;
      while(pos<size)
      { mark_pos+=SECTOR_SIZE;
        while(pos<mark_pos)
        { 
          if(bits.pd>bits.pe) goto error;	/* incompressible data */
          
          if(pos+1>=size) goto single_char;	/* cannot be hashed */
          
	  hash_cur=dbl_newhash(clusterd,pos,hash_tab,hash_hist,hash_mask);
          if(hash_cur>=pos) goto single_char;
   
          try_cn=try_count;
          max_match=0;
          do{                       /* longer offsets are not allowed */
            if(pos-hash_cur>=0x113F) break;
            if(clusterd[hash_cur]==clusterd[pos])   
            /* second chars are equal from hash function */
            {
              for(match=0;match<mark_pos-pos;match++)
                if(clusterd[hash_cur+match]!=clusterd[pos+match])break;
              if(match>max_match)  /* find maximal hash */
              { max_hash=hash_cur;max_match=match;
                if(match==mark_pos-pos) break;
              };
            };
            i=hash_cur;
          }while(--try_cn&&((hash_cur=hash_hist[i&hash_mask])<i));
          if(max_match<2) goto single_char;

            write_temp(&bits,&(clusterd[pos]),max_match,pos-max_hash,method);
            pos++;max_match--;
            i=max_match;if(pos+i+1>=size)i=size-pos-1;
            max_match-=i;	/* last char cannot be hashed */
            while(i--)
              dbl_newhash(clusterd,pos++,hash_tab,hash_hist,hash_mask);
            pos+=max_match;
            continue;
          
          single_char:
            write_byte(&bits,clusterd[pos++],method);

        }
        write_marker(&bits,method);
      }
      
      dblb_wrn(&bits,0,15);	/* flush last bits from bits.buf */
      if(bits.pd>bits.pe) goto error;
      FREE(hash_tab);
      FREE(hash_hist);
      return (((__u8*)bits.pd-(__u8*)clusterk)-1)/512+1;
      
      error:
        FREE(hash_tab);
        FREE(hash_hist);
        return -1;
      
    default: 
      /* sorry, other compression methods currently not available */
      return -1; 
  }
  
  return -1;
}

/* write a dmsdos cluster, compress before if possible;
   length is the number of used bytes, may be < SECTOR_SIZE*sectperclust only
   in the last cluster of a file;
   cluster must be allocated by allocate_cluster before if it is a new one;
   unable to write dir clusters;
   to avoid MDFAT level fragmentation, near_sector should be the sector no
   of the preceeding cluster;
   if ucflag==1 uncompressed write is forced (only for umsdos --linux-.---)
   if ucflag<0 raw write is forced with compressed size -ucflag (in bytes)
   if ucflag==2 simulate write is done (reserves space for the cluster in the
   filesystem but does not actually write it) 
   if ucflag==3 write compressed but don't use daemon - this is to guarantee
   that the data are on the disk when the function exits
   
   *********** This function is doublespace/drivespace specific ******
*/

#ifdef __KERNEL__

int dbl_write_cluster(struct super_block*sb,
                         unsigned char* clusterd, int length, int clusternr,
                         int near_sector, int ucflag)
{ int method;
  unsigned char* clusterk;
  int size;
  Mdfat_entry mde;
  int sektor;
  int i;
  int res;
  int ksize;
  struct buffer_head*bh;
  Dblsb*dblsb=MSDOS_SB(sb)->private_data;

  LOG_CLUST("DMSDOS dbl_write_cluster clusternr=%d length=%d near_sector=%d\n",
         clusternr,length,near_sector);

  /* are we deleting a cluster ? */
  if(clusterd==NULL||length==0)
  { free_cluster_sectors(sb,clusternr);
    return 0;
  }

  /* guess compression method if not already known */
  if(dblsb->s_comp==GUESS)
  { 
    printk("DMSDOS: write_cluster: guessing compression method...\n");
    { if(dblsb->s_cvf_version==DRVSP)
      { dblsb->s_comp=JM_0_0;
        /* no doubt here, there's only this one possible - so exit */
        /* goto guess_ok; Hmm ... really...???? better let it scan */
      }
      if(dblsb->s_cvf_version==DRVSP3)
      { dblsb->s_comp=SQ_0_0;
        /* that's only a default in case the scan routine finds nothing */
      }
      /* DBLSP: we know nothing, it can be DS_0_2 for dos 6.0 and DS_0_0
         for win95, let's scan it */
      for(i=2;i<=dblsb->s_max_cluster;++i)
      { dbl_mdfat_value(sb,i,NULL,&mde);
        /*if((mdfat&0xC0000000)==0x80000000)*/
        if(mde.flags==2)
        { bh=raw_bread(sb,mde.sector_minus_1+1);
          if(bh!=NULL)
          {
            res=CHL(bh->b_data);
            raw_brelse(sb,bh);
            if(res==DS_0_0){dblsb->s_comp=DS_0_0;goto guess_ok;}
            if(res==DS_0_1){dblsb->s_comp=DS_0_1;goto guess_ok;}
            if(res==DS_0_2){dblsb->s_comp=DS_0_2;goto guess_ok;}
            if(res==JM_0_0){dblsb->s_comp=JM_0_0;goto guess_ok;}
            if(res==JM_0_1){dblsb->s_comp=JM_0_1;goto guess_ok;}
            if(res==SQ_0_0){dblsb->s_comp=SQ_0_0;goto guess_ok;}
          }
        }
      }
      if(dblsb->s_comp==GUESS) /* still unknown ? */
      { printk("DMSDOS: could not guess compression method for CVF\n");
        dblsb->s_comp=UNCOMPRESSED;
      }
    }
    guess_ok:
    printk("DMSDOS: write_cluster: guessed 0x%08x.\n",dblsb->s_comp);
#ifdef GUESS_HACK
    if(dblsb->s_comp==SQ_0_0)
    { dblsb->s_comp=JM_0_1;
      printk("DMSDOS: guess_hack: guessed SQ-0-0 not supported, using JM-0-1 instead.\n");
    }
#endif
  }
  
  method=dblsb->s_comp; /* default compression method */
  
  size=(length-1)/512+1;
  if(size==1||(ucflag!=0&&ucflag!=3))method=UNCOMPRESSED; /* will not become smaller */
  
  if(method==UNCOMPRESSED) /* includes RAW and SIMULATE writes */
  { clusterk=clusterd;
    if(ucflag<0)
    { /* raw write of already compressed data (for dmsdosd/ioctl) */
      ksize=-ucflag/SECTOR_SIZE; /* must be n*512 for doublespace */
      mde.size_lo_minus_1=ksize-1;
      mde.size_hi_minus_1=size-1;
      mde.flags=2;
    }
    else
    { /* normal uncompressed write */
      /*mdfat=0xC0000000|((size-1)<<26)|((size-1)<<22);*/
      mde.size_lo_minus_1=size-1;
      mde.size_hi_minus_1=size-1;
      mde.flags=3;
      ksize=size;
    }
  }
  else
  { if((ucflag==3)?0:try_daemon(sb,clusternr,length,method))goto wr_uc;
    clusterk=(unsigned char*)MALLOC(size*SECTOR_SIZE);
    if(clusterk==NULL)
    { printk("DMSDOS: write_cluster: no memory for compression, writing uncompressed!\n");
     wr_uc:
      clusterk=clusterd;
      /*mdfat=0xC0000000|((size-1)<<26)|((size-1)<<22);*/
      mde.size_lo_minus_1=size-1;
      mde.size_hi_minus_1=size-1;
      mde.flags=3;
      method=UNCOMPRESSED;
      ksize=size;
    }
    else 
    { LOG_CLUST("DMSDOS: write_cluster: compressing...\n");
      ksize=dbl_compress(clusterk,clusterd,size,method,dblsb->s_cfaktor);
      LOG_CLUST("DMSDOS: write cluster: compressing finished\n");
      if(ksize<0)
      { /* compression failed */
        FREE(clusterk);
        clusterk=clusterd;
        /*mdfat=0xC0000000|((size-1)<<26)|((size-1)<<22);*/
        mde.size_lo_minus_1=size-1;
        mde.size_hi_minus_1=size-1;
        mde.flags=3;
        method=UNCOMPRESSED;
        ksize=size;
      }
      else
      { /*mdfat=0x80000000|((size-1)<<26)|((ksize-1)<<22);*/
        mde.size_lo_minus_1=ksize-1;
        mde.size_hi_minus_1=size-1;
        mde.flags=2;
      }
    }
  }
  
  LOG_CLUST("DMSDOS: write_cluster: call replace_existing_cluster\n");
  sektor=replace_existing_cluster(sb,clusternr,near_sector,&mde);
  LOG_CLUST("DMSDOS: write_cluster: replace_existing_cluster returned %d\n",
         sektor);
  if(sektor<0)res=-ENOSPC;
  else
  { res=0;
    if(ucflag!=2) /* no SIMULATE write */
    {
      for(i=0;i<ksize;++i)
      { bh=raw_getblk(sb,sektor+i);
        if(bh==NULL)res=-EIO;
        else
        {
          memcpy(bh->b_data,&(clusterk[SECTOR_SIZE*i]),SECTOR_SIZE);
          raw_set_uptodate(sb,bh,1);
          raw_mark_buffer_dirty(sb,bh,1);
          raw_brelse(sb,bh);
        }
      }
    }
  }
  
  if(method!=UNCOMPRESSED)FREE(clusterk);
  
  return res;
}

/* this is really a shame that this functions must exist :-(( */
struct wait_queue * write_wait = NULL;
int write_lock=0;
void lock_write(void)
{ while(write_lock)sleep_on(&write_wait);
  write_lock=1;
}
void unlock_write(void)
{ write_lock=0;
  wake_up(&write_wait);
}
    
/* write a dmsdos cluster, compress before if possible;
   length is the number of used bytes, may be < SECTOR_SIZE*sectperclust only
   in the last cluster of a file;
   cluster must be allocated by allocate_cluster before if it is a new one;
   unable to write dir clusters;
   to avoid MDFAT level fragmentation, near_sector should be the sector no
   of the preceeding cluster;
   if ucflag==1 uncompressed write is forced (only for umsdos --linux-.---)
   if ucflag<0 raw write is forced with compressed size -ucflag (in bytes)
   if ucflag==2 simulate write is done (reserves space for the cluster in the
   filesystem but does not actually write it) 
   length==0 or clusterd==NULL means remove the cluster
   if ucflag==3 perform like ucflag==0 but don't use the daemon
      
   *********** This function is a generic wrapper ******
*/

int dmsdos_write_cluster(struct super_block*sb,
                         unsigned char* clusterd, int length, int clusternr,
                         int near_sector, int ucflag)
{ int ret;
  Dblsb*dblsb=MSDOS_SB(sb)->private_data;

  LOG_CLUST("DMSDOS: write_cluster clusternr=%d length=%d near_sector=%d\n",
         clusternr,length,near_sector);

  /* ensure the daemon doesn't use old data and overwrites our data again 
     but don't do this when called by the daemon itself :-/ uuhhh deadlock */
  /* also don't do it for simulated writes - they change nothing.... */
  if(ucflag>=0&&ucflag!=2)remove_from_daemon_list(sb,clusternr);

  /* check whether using the daemon is not desired due to speedup bits */
  if(ucflag==0&&(dmsdos_speedup&SP_USE_DAEMON)==0)ucflag=3;
  
  /* we need locking here though we shouldn't :-( */
  lock_write();
  
  switch(dblsb->s_cvf_version)
  { case DBLSP:
    case DRVSP:
    case DRVSP3:
      ret=dbl_write_cluster(sb,clusterd,length,clusternr,near_sector,
                            ucflag);
      break;
    case STAC3:
    case STAC4:
      ret=stac_write_cluster(sb,clusterd,length,clusternr,near_sector,
                             ucflag);
      break;
    default:
      printk("DMSDOS: write_cluster: illegal cvf_version flag!\n");
      ret=-EIO;
  }
  
  unlock_write();
  
  return ret;
}

#endif /*__KERNEL__*/
