#include <linux/sched.h>
#include <linux/ctype.h>
#include <linux/major.h>
#include <linux/blkdev.h>
#include <linux/fs.h>
#include <linux/stat.h>
#include <linux/locks.h>
#include <asm/segment.h>
#include <linux/mm.h>
#include <linux/malloc.h>
#include <linux/string.h>
#include <linux/msdos_fs.h>
#include <linux/errno.h>
#include <linux/kernel.h>
#include <linux/shm.h>
#include <linux/mman.h>
#include <asm/system.h>
#include "dmsdos.h"

Acache mdfat[MDFATCACHESIZE];
Acache dfat[DFATCACHESIZE];
Acache bitfat[BITFATCACHESIZE];

extern unsigned long int dmsdos_speedup;

struct wait_queue * mdfatwait=NULL;
int mdfatlock=0;
void lock_mdfat(void)
{ while(mdfatlock)sleep_on(&mdfatwait);
  mdfatlock=1;
}
void unlock_mdfat(void)
{ mdfatlock=0;
  wake_up(&mdfatwait);
}

struct wait_queue * dfatwait=NULL;
int dfatlock=0;
void lock_dfat(void)
{ while(dfatlock)sleep_on(&dfatwait);
  dfatlock=1;
}
void unlock_dfat(void)
{ dfatlock=0;
  wake_up(&dfatwait);
}

struct wait_queue * bitfatwait=NULL;
int bitfatlock=0;
void lock_bitfat(void)
{ while(bitfatlock)sleep_on(&bitfatwait);
  bitfatlock=1;
}
void unlock_bitfat(void)
{ bitfatlock=0;
  wake_up(&bitfatwait);
}

void read_mdfat_sector(struct super_block*sb,int area,int merk_i)
{
  LOG_MDFAT("DMSDOS: reading dbl_mdfat sector %d\n",area);
  
  if(mdfat[merk_i].a_buffer!=NULL)
          raw_brelse(mdfat[merk_i].a_sb,mdfat[merk_i].a_buffer);
  if((mdfat[merk_i].a_buffer=raw_bread(sb,area))==NULL)
  { panic("DMSDOS: unable to read mdfat\n");
  }
  mdfat[merk_i].a_area=area;
  mdfat[merk_i].a_time=CURRENT_TIME;
  mdfat[merk_i].a_acc=0;
  mdfat[merk_i].a_sb=sb;
}

void read_dfat_sector(struct super_block*sb,int area,int merk_i)
{
  LOG_DFAT("DMSDOS: reading dbl_fat sector %d\n",area);
  
  if(dfat[merk_i].a_buffer!=NULL)
          raw_brelse(dfat[merk_i].a_sb,dfat[merk_i].a_buffer);
  if((dfat[merk_i].a_buffer=raw_bread(sb,area))==NULL)
  { panic("DMSDOS: unable to read dfat\n");
  }
  dfat[merk_i].a_area=area;
  dfat[merk_i].a_time=CURRENT_TIME;
  dfat[merk_i].a_acc=0;
  dfat[merk_i].a_sb=sb;
  
}

void read_bitfat_sector(struct super_block*sb,int area,int merk_i)
{
  LOG_BITFAT("DMSDOS: reading bitfat sector %d\n",area);
  
  if(bitfat[merk_i].a_buffer!=NULL)
         raw_brelse(bitfat[merk_i].a_sb,bitfat[merk_i].a_buffer);
  if((bitfat[merk_i].a_buffer=raw_bread(sb,area))==NULL)
  { panic("DMSDOS: unable to read bitfat\n");
  }
  bitfat[merk_i].a_area=area;
  bitfat[merk_i].a_time=CURRENT_TIME;
  bitfat[merk_i].a_acc=0;
  bitfat[merk_i].a_sb=sb;
}

void u_dumpcache(Acache*c)
{ printk("area=%d time=%ld acc=%d buffer=%p\n",c->a_area,c->a_time,
         c->a_acc,c->a_buffer);
}

void dumpcache(void)
{ int i;
  
  printk("DMSDOS: mdfat cache:\n");
  for(i=0;i<MDFATCACHESIZE;++i)u_dumpcache(&(mdfat[i]));
  printk("DMSDOS: dfat cache:\n");
  for(i=0;i<DFATCACHESIZE;++i)u_dumpcache(&(dfat[i]));
  printk("DMSDOS: bitfat cache:\n");
  for(i=0;i<BITFATCACHESIZE;++i)u_dumpcache(&(bitfat[i]));
}

void u_free_idle_cache(Acache*c)
{ if(c->a_buffer!=NULL&&c->a_time-CURRENT_TIME>MAX_CACHE_TIME)
  { raw_brelse(c->a_sb,c->a_buffer);
    c->a_buffer=NULL;
    c->a_time=0;
    c->a_acc=0;
  }
}

void free_idle_cache(void)
{ int i;
  
  lock_mdfat();
  for(i=0;i<MDFATCACHESIZE;++i)u_free_idle_cache(&(mdfat[i]));
  unlock_mdfat();
  lock_dfat();
  for(i=0;i<DFATCACHESIZE;++i)u_free_idle_cache(&(dfat[i]));
  unlock_dfat();
  lock_bitfat();
  for(i=0;i<BITFATCACHESIZE;++i)u_free_idle_cache(&(bitfat[i]));
  unlock_bitfat();
  
  /* handle cluster cache */
  free_idle_ccache();
}

void dbl_mdfat_value(struct super_block* sb,int clusternr,
                     Mdfat_entry*new,
                     Mdfat_entry*mde)
{ int i;
  int area;
  int pos;
  int offset;
  unsigned long min_time;
  unsigned int min_acc;
  int merk_i,merk_i2,merk_i3;
  unsigned char * pp;
  int res;
  unsigned char mdfat_raw_field[5];
  Dblsb*dblsb=MSDOS_SB(sb)->private_data;
    
  if(dblsb->s_cvf_version>=STAC3)
  { /* stacker */ 
    if(dblsb->s_16bitfat)pos=clusternr*4;
    else pos=clusternr*3;
    area=pos/SECTOR_SIZE;
    offset=pos%SECTOR_SIZE;
    area=(area/6)*9+(area%6)+3+dblsb->s_fatstart; /* yes!!! */
  }

  else
  { /* dblspace/drvspace */
    if(dblsb->s_sectperclust>16)
      /* drvspace 3 */
      pos=(dblsb->s_dcluster+clusternr)*5
          +((dblsb->s_dcluster+clusternr)/102)*2
          +512*dblsb->s_mdfatstart;
    else  
      pos=(dblsb->s_dcluster+clusternr)*4+512*dblsb->s_mdfatstart;
    area=pos/SECTOR_SIZE;
    offset=(pos%SECTOR_SIZE);
  }                           
  
  lock_mdfat();
  
  merk_i2=-1; /* never free this cached sector */
  
mdfat_redo:
  
  min_time=mdfat[0].a_time;
  min_acc=mdfat[0].a_acc;
  merk_i=0;
  if(merk_i2==0)
  { min_time=mdfat[1].a_time;
    min_acc=mdfat[1].a_acc;
    merk_i=1;
  }
  /* find area and dev in cache */
  for(i=0;i<MDFATCACHESIZE;++i)
  { if( ( mdfat[i].a_time<min_time||
            (mdfat[i].a_time==min_time&&mdfat[i].a_acc<min_acc)
        ) &&merk_i2!=i)    
    { min_time=mdfat[i].a_time;
      min_acc=mdfat[i].a_acc;
      merk_i=i;
    }
    if(mdfat[i].a_buffer!=NULL&&area==mdfat[i].a_area&&sb->s_dev==mdfat[i].a_sb->s_dev)
    { /* found */
      if(mdfat[i].a_time==CURRENT_TIME)++mdfat[i].a_acc;
      else
      { mdfat[i].a_time=CURRENT_TIME;
        mdfat[i].a_acc=0;
      }
      merk_i=i;
      goto mdfat_okay;      /************* goto ************************/
    }
  }
  /* merk_i = least recently used entry number */
  read_mdfat_sector(sb,area,merk_i);

mdfat_okay:

  if(dblsb->s_cvf_version>=STAC3)
  { /* stacker */
    int anzb=3;
    if(dblsb->s_16bitfat)anzb=4;
  
    /* read 2. sector if necessary */
    if(offset+anzb-1>511&&merk_i2==-1)
    { merk_i2=merk_i;
      ++area;
      goto mdfat_redo;
    }
    if(offset+anzb-1>511)
    { merk_i3=merk_i;
      merk_i=merk_i2;
      merk_i2=merk_i3;
    }
    else merk_i2=merk_i;
  
    /* copy data in mdfat_raw_field (anzb byte) */
    mdfat_raw_field[3]=0; /* in case only 3 bytes to read */
    for(i=0;i<anzb;++i)
    { if(i+offset<512)
        mdfat_raw_field[i]=mdfat[merk_i].a_buffer->b_data[i+offset];
      else
        mdfat_raw_field[i]=mdfat[merk_i2].a_buffer->b_data[i+offset-512];
    }
  
    /* setup mde */
    mde->sector_minus_1=CHS(mdfat_raw_field)+((mdfat_raw_field[3]&0x3f)<<16)-1;
    mde->unknown=(mdfat_raw_field[2])&0xf0; /* same as flags here */
    mde->size_lo_minus_1=(mdfat_raw_field[2]&0xf)+((mdfat_raw_field[3]>>2)&0x30);
    mde->size_hi_minus_1=mde->size_lo_minus_1;  /* for compatibility */
    mde->flags=(mdfat_raw_field[2])&0xf0;   /* unshifted like in sd4_c.cc */
    /* set used and compressed bits in flags (for compatibility) */
    /* Hi Pavel: Is there a bug here? sector seems to be sometimes 1 for
       empty stacker mdfat slots. I think it should be 0 ? This showed up in 
       the new fs checking code as dead sectors. Hmm.... Should we tolerate
       0 and 1 here ? But then the stacker fs check complains later... */
    /* Hi Frank, they are normal deleted clusters, stored for DOS undel.
       They should be deleted automaticaly, when free space becomes low.
       At this moment is best to ignore them in your fat simple check,
       they are counted by stac simple check */
    if(mde->sector_minus_1+1)mde->flags|=2; /*set 'used' flag*/
    switch(mde->flags&0xa0)
    { case 0x80:
      case 0x00:
        if(mde->size_lo_minus_1+1==dblsb->s_sectperclust)
          mde->flags|=1;
        break;
      case 0x20:
        mde->flags|=1;
        break;
      default:
        /* correct length for compatibility */
        mde->size_hi_minus_1=dblsb->s_sectperclust-1;
    }
    
    LOG_MDFAT("DMSDOS: dbl_mdfat_value : cluster %u\n",
           (unsigned)clusternr);
    LOG_MDFAT("DMSDOS:    sector %u len %u flags 0x%X raw 0x%02X 0x%02X 0x%02X 0x%02X\n",
           (unsigned)(mde->sector_minus_1+1),(unsigned)(mde->size_lo_minus_1+1),
           (unsigned)mde->flags,
           (unsigned)(mdfat_raw_field[0]),(unsigned)(mdfat_raw_field[1]),
           (unsigned)(mdfat_raw_field[2]),(unsigned)(mdfat_raw_field[3]));
    LOG_MDFAT("DMSDOS:    pos %u area %u offset %u\n",
           (unsigned)pos,(unsigned)area,(unsigned)offset);
  
    /* if new!=NULL, setup anzb byte from new in mdfat_raw_field */
    if(new)
    { mdfat_raw_field[0]=(new->sector_minus_1+1);
      /* unknown bits ignored */
      mdfat_raw_field[1]=(new->sector_minus_1+1)>>8;
      mdfat_raw_field[3]=((new->sector_minus_1+1)>>16)&0x3f;
      mdfat_raw_field[3]|=(new->size_lo_minus_1<<2)&0xc0;
      mdfat_raw_field[2]=new->size_lo_minus_1&0x0f;
      mdfat_raw_field[2]|=new->flags&0xf0; /* unshifted like in sd4_c.cc */
    
      /* write back mdfat_raw_entry */
      for(i=0;i<anzb;++i)
      { if(i+offset<512)
          mdfat[merk_i].a_buffer->b_data[i+offset]=mdfat_raw_field[i];
        else
          mdfat[merk_i2].a_buffer->b_data[i+offset-512]=mdfat_raw_field[i];
      }
    
      /* mark buffer dirty (both if necessary) */
      raw_mark_buffer_dirty(sb,mdfat[merk_i].a_buffer,1);
      if(merk_i!=merk_i2)
          raw_mark_buffer_dirty(sb,mdfat[merk_i2].a_buffer,1);
      /* write second mdfat if it exists */
      if(dblsb->s_2nd_fat_offset)
      { struct buffer_head* bh;
      
        bh=raw_getblk(sb,
                      mdfat[merk_i].a_area+dblsb->s_2nd_fat_offset);
        if(bh==NULL)panic("unable to read second mdfat");
        memcpy(bh->b_data,mdfat[merk_i].a_buffer->b_data,SECTOR_SIZE);
        raw_set_uptodate(sb,bh,1);
        raw_mark_buffer_dirty(sb,bh,1);
        raw_brelse(sb,bh);
        if(merk_i!=merk_i2)
        { bh=raw_getblk(sb,
                 mdfat[merk_i2].a_area+dblsb->s_2nd_fat_offset);
          if(bh==NULL)panic("unable to read second mdfat");
          memcpy(bh->b_data,mdfat[merk_i2].a_buffer->b_data,SECTOR_SIZE);
          raw_set_uptodate(sb,bh,1);
          raw_mark_buffer_dirty(sb,bh,1);
          raw_brelse(sb,bh);                              
        }
      }
    
    }
  
    unlock_mdfat();
    return;
  }

  if(dblsb->s_sectperclust<=16)
  { /* normal doublesp/drivesp */
    pp=&(mdfat[merk_i].a_buffer->b_data[offset]);
    res=CHL(pp);
    mde->sector_minus_1=res&0x1fffff;
    mde->unknown=(res&0x00200000)>>21;
    mde->size_lo_minus_1=(res&0x03c00000)>>22;
    mde->size_hi_minus_1=(res&0x3c000000)>>26;
    mde->flags=((res&0xC0000000)>>30)&3;
    if(new)
    { res=new->sector_minus_1;res&=0x1fffff;
      /* unknown bit ??? don't know... set to zero */
      res|=new->size_lo_minus_1<<22;res&=0x03ffffff;
      res|=new->size_hi_minus_1<<26;res&=0x3fffffff;
      res|=new->flags<<30;
      pp[0]=res;pp[1]=res>>8;pp[2]=res>>16;pp[3]=res>>24;
      raw_mark_buffer_dirty(sb,mdfat[merk_i].a_buffer,1);
    }
    unlock_mdfat();
    return;
  }

  /* drivspace 3 mdfat access  _experimental_ but seems to work */

  /* read 2. sector if necessary */
  if(offset+4>511&&merk_i2==-1)
  { merk_i2=merk_i;
    ++area;
    goto mdfat_redo;
  }
  if(offset+4>511)
  { merk_i3=merk_i;
    merk_i=merk_i2;
    merk_i2=merk_i3;
  }
  else merk_i2=merk_i;
  
  /* copy data in mdfat_raw_field (5 byte) */
  for(i=0;i<5;++i)
  { if(i+offset<512)
      mdfat_raw_field[i]=mdfat[merk_i].a_buffer->b_data[i+offset];
    else
      mdfat_raw_field[i]=mdfat[merk_i2].a_buffer->b_data[i+offset-512];
  }
  
  /* setup mde */
  mde->sector_minus_1=CHL(mdfat_raw_field)&0xffffff;
  mde->unknown=(CHL(mdfat_raw_field)&0x3000000)>>24;
  mde->size_lo_minus_1=(mdfat_raw_field[3]>>2)&0x3f;
  mde->size_hi_minus_1=mdfat_raw_field[4]&0x3f;
  mde->flags=(mdfat_raw_field[4]>>6)&3;
  
  /* if new!=NULL, setup 5 byte from new in mdfat_raw_field */
  if(new)
  { mdfat_raw_field[0]=new->sector_minus_1&0xffffff;
    /* unknown bits set to zero */
    mdfat_raw_field[1]=new->sector_minus_1>>8;
    mdfat_raw_field[2]=new->sector_minus_1>>16;
    mdfat_raw_field[3]=(new->sector_minus_1>>24)&3;
    mdfat_raw_field[3]|=new->size_lo_minus_1<<2;
    mdfat_raw_field[4]=new->size_hi_minus_1&0x3f;
    mdfat_raw_field[4]|=new->flags<<6;
    
    /* write back mdfat_raw_entry */
    for(i=0;i<5;++i)
    { if(i+offset<512)
        mdfat[merk_i].a_buffer->b_data[i+offset]=mdfat_raw_field[i];
      else
        mdfat[merk_i2].a_buffer->b_data[i+offset-512]=mdfat_raw_field[i];
    }
    
    /* mark buffer dirty (both if necessary) */
    raw_mark_buffer_dirty(sb,mdfat[merk_i].a_buffer,1);
    if(merk_i!=merk_i2)
          raw_mark_buffer_dirty(sb,mdfat[merk_i2].a_buffer,1);
    
  }
  
  unlock_mdfat();
  return;
}

int dbl_mdfat_cluster2sector(struct super_block*sb,int clusternr)
{ Mdfat_entry mde;
  Dblsb*dblsb=MSDOS_SB(sb)->private_data;
  
  if(clusternr==0)return dblsb->s_rootdir;
  dbl_mdfat_value(sb,clusternr,NULL,&mde);
  return mde.sector_minus_1+1;
}

int dbl_fat_nextcluster(struct super_block*sb,int clusternr,int*new)
{ int i;
  int area;
  int pos;
  int offset;
  unsigned long min_time;
  unsigned int min_acc;
  int merk_i;
  int res;
  int newval;
  int merk_i2;
  int merk_i3;
  int offset2;
  Dblsb*dblsb=MSDOS_SB(sb)->private_data;
    
  merk_i2=-1; /* never free this cached sector */

/*
  pos= dblsb->s_fatstart*512;
  pos+= dblsb->s_16bitfat ? clusternr*2 : clusternr*3/2;
       
  area=pos/SECTOR_SIZE;
  offset=pos%SECTOR_SIZE;                           
*/

  pos= dblsb->s_16bitfat ? clusternr<<1/**2*/ : (clusternr*3)>>1/*/2*/;
  area=pos>>SECTOR_BITS/*pos/SECTOR_SIZE*/;
  offset=pos&(SECTOR_SIZE-1)/*pos%SECTOR_SIZE*/;
  /* adapt area for stacker */
  if(dblsb->s_cvf_version>=STAC3)area=(area/3)*9+(area%3);
  
  area+=dblsb->s_fatstart;
  
  lock_dfat();
  
dfat_retry:
  
  min_time=dfat[0].a_time;
  min_acc=dfat[0].a_acc;
  merk_i=0;
  if(merk_i2==0)
  { min_time=dfat[1].a_time;
    min_acc=dfat[1].a_acc;
    merk_i=1;
  }
  /* find area and dev in cache */
  for(i=0;i<DFATCACHESIZE;++i)
  { if( ( dfat[i].a_time<min_time ||
           (dfat[i].a_time==min_time&&dfat[i].a_acc<min_acc) 
        ) &&merk_i2!=i)
    { min_time=dfat[i].a_time;
      min_acc=dfat[i].a_acc;
      merk_i=i;
    }
    if(dfat[i].a_buffer!=NULL&&area==dfat[i].a_area&&sb->s_dev==dfat[i].a_sb->s_dev)
    { /* found */
      if(dfat[i].a_time==CURRENT_TIME)++dfat[i].a_acc;
      else
      { dfat[i].a_time=CURRENT_TIME;
        dfat[i].a_acc=0;
      }
      merk_i=i;
      goto dfat_okay;
    }
  }
  /* merk_i = least recently used entry */
  read_dfat_sector(sb,area,merk_i);
  
dfat_okay:
  if(offset==511&&merk_i2==-1)
  { /* overlap, must read following sector also */
    merk_i2=merk_i;
    ++area;
    goto dfat_retry;
  }
  if(offset==511)
  { /* we have just read the second sector (overlap case) */
    merk_i3=merk_i;
    merk_i=merk_i2; /* wiederherstellen */
    merk_i2=merk_i3; /* folgesektor */
    offset2=0;
  }
  else
  { /* normal */
    merk_i2=merk_i;
    offset2=offset+1;
  }

LOG_DFAT("DMSDOS FAT lookup: area=%d merk_i=%d merk_i2=%d offset=%d offset2=%d\n",
       area,merk_i,merk_i2,offset,offset2);
LOG_DFAT("DMSDOS FAT lookup: cluster=%d value(low=%d high=%d)\n",
       clusternr,
       dfat[merk_i].a_buffer->b_data[offset],
       dfat[merk_i2].a_buffer->b_data[offset2]); 

  res=dfat[merk_i].a_buffer->b_data[offset];
  res&=0xff; /* grmpf... sign problems !!!! this is brutal but it works */
  res|=dfat[merk_i2].a_buffer->b_data[offset2]<<8;
  res&=0xffff;
  
  if(new)
  { if(dblsb->s_16bitfat) newval=*new&0xFFFF;
    else
    { if(clusternr&1) newval=(res&0xF) | ((*new&0xFFF)<<4);
      else newval=(res&0xF000) | (*new&0xFFF);
    }
    dfat[merk_i].a_buffer->b_data[offset]=newval;
    dfat[merk_i2].a_buffer->b_data[offset2]=newval>>8;
    raw_mark_buffer_dirty(sb,dfat[merk_i].a_buffer,1);
    if(merk_i!=merk_i2)
           raw_mark_buffer_dirty(sb,dfat[merk_i2].a_buffer,1);
    /* write second fat if it exists */
      if(dblsb->s_2nd_fat_offset)
      { struct buffer_head* bh;
      
        bh=raw_getblk(sb,
               dfat[merk_i].a_area+dblsb->s_2nd_fat_offset);
        if(bh==NULL)panic("unable to read second fat");
        memcpy(bh->b_data,dfat[merk_i].a_buffer->b_data,SECTOR_SIZE);
        raw_set_uptodate(sb,bh,1);
        raw_mark_buffer_dirty(sb,bh,1);
        raw_brelse(sb,bh);
        if(merk_i!=merk_i2)
        { bh=raw_getblk(sb,
               dfat[merk_i2].a_area+dblsb->s_2nd_fat_offset);
          if(bh==NULL)panic("unable to read second fat");
          memcpy(bh->b_data,dfat[merk_i2].a_buffer->b_data,SECTOR_SIZE);
          raw_set_uptodate(sb,bh,1);
          raw_mark_buffer_dirty(sb,bh,1);
          raw_brelse(sb,bh);                              
        }
      }

  }
  unlock_dfat();
  if(dblsb->s_16bitfat)return res>=0xFFF7 ? -1 : res;
  if(clusternr&1)res>>=4; else res&=0xfff;
  return res>=0xff7 ? -1 : res;
}

int dbl_bitfat_value(struct super_block*sb,int sektornr,int*new)
{ int i;
  int area;
  int pos;
  int offset;
  unsigned long min_time;
  unsigned int min_acc;
  int merk_i;
  unsigned char * pp;
  int bitmask,shiftval;
  int res;
  int newval;
  Dblsb*dblsb=MSDOS_SB(sb)->private_data;
    
  if(sektornr<dblsb->s_datastart)return -1;
  if(sektornr>dblsb->s_dataend)return -1;
  if(dblsb->s_cvf_version==STAC3)
  { /* stacker 3 */
    pos=((sektornr-dblsb->s_datastart)>>3);
    offset=pos&(SECTOR_SIZE-1)/*pos%SECTOR_SIZE*/;
    area=(pos>>SECTOR_BITS)/*pos/SECTOR_SIZE*/+dblsb->s_mdfatstart; /* misused */
    shiftval=((sektornr-dblsb->s_datastart)&7);
    bitmask=0x1;
  } /* stacker 4 */
  else if(dblsb->s_cvf_version==STAC4)
  { pos=((sektornr-dblsb->s_datastart)>>2);
    offset=pos&(SECTOR_SIZE-1)/*pos%SECTOR_SIZE*/;
    area=(pos>>SECTOR_BITS)/*pos/SECTOR_SIZE*/+dblsb->s_mdfatstart; /* misused */
    shiftval=((sektornr-dblsb->s_datastart)&3)<<1/**2*/;
    bitmask=0x3;
  }
  else
  { /* dblspace/drvspace */
    pos=SECTOR_SIZE+((sektornr-dblsb->s_datastart)>>4)*2;
    area=pos>>SECTOR_BITS/*pos/SECTOR_SIZE*/;
    offset=pos&(SECTOR_SIZE-1)/*(pos%SECTOR_SIZE)*/;
    bitmask=0x8000>>((sektornr-dblsb->s_datastart)&15);
    shiftval=0; /* unused - avoids warning */ 
  }
  
  lock_bitfat();
  
  min_time=bitfat[0].a_time;
  min_acc=bitfat[0].a_acc;
  merk_i=0;
  /* find area and dev in cache */
  for(i=0;i<BITFATCACHESIZE;++i)
  { if(bitfat[i].a_time<min_time ||
        (bitfat[i].a_time==min_time&&bitfat[i].a_acc<min_acc) )
    { min_time=bitfat[i].a_time;
      min_acc=bitfat[i].a_acc;
      merk_i=i;
    }
    if(bitfat[i].a_buffer!=NULL&&area==bitfat[i].a_area&&
       sb->s_dev==bitfat[i].a_sb->s_dev)
    { /* found */
      if(bitfat[i].a_time==CURRENT_TIME)++bitfat[i].a_acc;
      else
      { bitfat[i].a_time=CURRENT_TIME;
        bitfat[i].a_acc=0;
      }
      merk_i=i;
      goto bitfat_okay;      /************* goto ************************/
    }
  }
  /* merk_i = least recently used entry number */
  read_bitfat_sector(sb,area,merk_i);

bitfat_okay:
  if(dblsb->s_cvf_version<STAC3)
  { /* dblspace/drvspace */
    pp=&(bitfat[merk_i].a_buffer->b_data[offset]);
    res=CHS(pp);
    if(new)
    { newval= (*new) ? (res|bitmask) : (res&~bitmask);
      pp[0]=newval;pp[1]=newval>>8;
      raw_mark_buffer_dirty(sb,bitfat[merk_i].a_buffer,1);
    }
    res=(res&bitmask) ? 1 : 0;
  }
  else if(dblsb->s_cvf_version>=STAC3)
  { /* stacker 3 / 4 */
    res=(bitfat[merk_i].a_buffer->b_data[offset]>>shiftval)&bitmask;
    if(new)
    { bitfat[merk_i].a_buffer->b_data[offset]&=~(bitmask<<shiftval);
      bitfat[merk_i].a_buffer->b_data[offset]|=(*new&bitmask)<<shiftval;
      raw_mark_buffer_dirty(sb,bitfat[merk_i].a_buffer,1);
    }
  }
  unlock_bitfat();
  return res;
}

int dblspace_fat_access(struct super_block*sb, int clusternr, int newval)
{ int cl;

  cl=dbl_fat_nextcluster(sb,clusternr,NULL);
  
  if(newval==-1) return cl;
  if(newval==0)delete_cache_cluster(sb,clusternr);
  dbl_fat_nextcluster(sb,clusternr,&newval);
  return cl;
}

int dblspace_bmap(struct inode*inode, int block)
{ printk("DMSDOS bmap called, unsupported!\n");
  return -EIO;
}

int dblspace_smap(struct inode*inode, int block)
{ int cluster;
  int sect_offs;
  Dblsb*dblsb=MSDOS_SB(inode->i_sb)->private_data;
  
  cluster=block/dblsb->s_sectperclust;
  sect_offs=block%dblsb->s_sectperclust;
  
  LOG_REST("DMSDOS: smap called, block=%d cluster_offs=%d sector_offs=%d\n",
          block,cluster,sect_offs);
  
  if (inode->i_ino == MSDOS_ROOT_INO || (S_ISDIR(inode->i_mode) &&
     !MSDOS_I(inode)->i_start)) 
  {
     if (block >= MSDOS_SB(inode->i_sb)->dir_entries >> MSDOS_DPS_BITS) 
     { LOG_REST("DMSDOS: smap: root dir beyond end, returning 0\n");
       return 0;
     }
     LOG_REST("DMSDOS: smap: root dir, returning %d\n",block+FAKED_ROOT_DIR_OFFSET);
     return block+FAKED_ROOT_DIR_OFFSET;
  }
  
  if (!(cluster = get_cluster(inode,cluster))) 
  { LOG_REST("DMSDOS: smap: get_cluster returned 0\n");
    return 0;
  }
  LOG_REST("DMSDOS: smap: returning vsector(cluster=%d sector=%d)\n",
          cluster,sect_offs);
  return ((cluster-2)*dblsb->s_sectperclust)+sect_offs+FAKED_DATA_START_OFFSET;
}

/* clusternr is absolute, not relative to inode */
void dblspace_zero_new_cluster(struct inode*inode, int clusternr)
{ /* we decide upon the inode whether this is a dir cluster */
  struct super_block*sb=inode->i_sb;
  Dblsb*dblsb=MSDOS_SB(sb)->private_data;
  Cluster_head*ch;
  
  if(!S_ISDIR(inode->i_mode))
  { /* we just throw the cluster away if it has an mdfat entry */
    delete_cache_cluster(sb,clusternr);
  }
  else
  { /* it may be in cache, so ... */
    ch=ch_read(sb,clusternr);
    if(ch==NULL)
    { printk("DMSDOS: zero_new_cluster: ch_read failed???\n");
      return;
    }
    memset(ch->c_data,0,dblsb->s_sectperclust*SECTOR_SIZE);
    if(DIR_MAY_BE_SHORT(dblsb))ch->c_length=SECTOR_SIZE;
    else ch->c_length=dblsb->s_sectperclust*SECTOR_SIZE;
    ch_dirty(ch,0,DIR_MAY_BE_COMPRESSED(dblsb)?0:1);
  }
}
