#ifndef _DMSDOS_H
#define _DMSDOS_H

/*
 * The DMSDOS filesystem constants/structures
 */

#include <linux/fs.h>
#include <linux/stat.h>
#include <linux/fd.h>
#include "dmsdos-config.h"

#define DMSDOS_MAJOR      0
#define DMSDOS_MINOR      9
#define DMSDOS_ACT_REL    0
#define DMSDOS_COMP_REL   0
#define DMSDOS_PL         "1"
#define DMSDOS_EXTRA      "(alpha test version)"

#define DMSDOS_VERSION ((DMSDOS_MAJOR<<16)|(DMSDOS_MINOR<<8)|DMSDOS_ACT_REL)
#define DMSDOS_LOWEST_COMPATIBLE_VERSION ((DMSDOS_MAJOR<<16)|(DMSDOS_MINOR<<8)|DMSDOS_COMP_REL)
#define DMSDOS_VLT "pl" DMSDOS_PL DMSDOS_EXTRA

/* known compression methods */
#define DS_0_0 0x00005344
#define DS_0_1 0x01005344
#define DS_0_2 0x02005344
#define JM_0_0 0x00004D4A
/* drivespace 3 high compression */ 
#define JM_0_1 0x01004D4A
/* drivespace 3 ultra compression */
#define SQ_0_0 0x00005153
/* stacker 3 compression (no header) */
#define SD_3 0x00000000
/* stacker 4 compression */
#define SD_4 0x00000081

/* other defines for options */
#define READ_ONLY -1
#define UNCOMPRESSED -2
#define GUESS -3

typedef struct
{ unsigned long free_sectors;
  unsigned long used_sectors;
  unsigned long max_hole;
  unsigned long free_clusters;
  unsigned long used_clusters;
  unsigned long lost_clusters;
  unsigned long sectors_lo;
  unsigned long sectors_hi;
  unsigned long compressed_clusters;
  unsigned long uncompressed_clusters;
} Dblstat;                                                            

typedef struct
{ long sector_minus_1;
  short size_lo_minus_1;
  short size_hi_minus_1;
  short unknown; /* some bits I don't know to handle.... */
  short flags; /* 00...0uc - u=used, c=compressed */ 
} Mdfat_entry;

/* flag values */
#define D_EMPTY       0
#define D_VALID       1    /* entry is valid -> cluster to be compressed */
#define D_IN_D_ACTION 2    /* is being compressed by daemon */
#define D_OVERWRITTEN 3    /* has been overwritten by dmsdos while daemon
                            is compressing it -> throw away the result from
                            the daemon */
#ifdef __KERNEL__
#ifdef USE_XMALLOC
void* xmalloc(unsigned long);
void xfree(void*);
#define MALLOC(x) xmalloc(x)
#define FREE(x) xfree(x)
#else
#ifdef USE_VMALLOC
#include<linux/mm.h>
#ifdef __FOR_KERNEL_2_1_76
#include<linux/vmalloc.h>
#endif
#define MALLOC(x) vmalloc(x)
#define FREE(x) vfree(x)
#else
#include<linux/malloc.h>
#define MALLOC(x) kmalloc(x,GFP_KERNEL)
#define FREE(x) kfree(x)
#endif
#endif
#endif

/* this must be known outside the kernel too */
typedef struct {
  int s_dcluster;/*[45-46]*/
  int s_mdfatstart;/*[36-37]+1*/
  int s_fatstart;/*[39-40]+[14-15]*/ 
  int s_rootdir;/*[41-42]+[39-40]*/
  int s_rootdiranzentry;
  int s_sectperclust;
  int s_spc_bits;
  int s_16bitfat;
  int s_datastart;
  int s_dataend;
  int s_comp;
  int s_bootblock;/*[39-40]*/
  int s_cfaktor;
  int s_full;
  int s_max_cluster;
  int s_max_cluster2;
  int s_cvf_version;   /* dblsp/drvsp/drvsp3/stac3/stac4 */
  int s_2nd_fat_offset;
  int s_lastnear;
  int s_lastbig;
  int s_free_sectors;
  struct wait_queue*mdfat_alloc_wait;
  int mdfat_alloc_locked;
} Dblsb;

#define DBLSP   0
#define DRVSP   1
#define DRVSP3  2
#define STAC3   3
#define STAC4   4

/* cvf version capabilities - boolean values */
#define DIR_MAY_BE_SHORT(dblsb)       (dblsb->s_cvf_version==DRVSP3)
#define DIR_MAY_BE_COMPRESSED(dblsb)  (dblsb->s_cvf_version>=DRVSP3&&(dmsdos_speedup&SP_NO_DIR_COMPR)==0)
#define UMOUNT_UCFLAG ((dmsdos_speedup&SP_NO_UNMOUNT_COMPR)?1:3)

typedef struct {
  struct buffer_head * a_buffer;
  unsigned int a_area;
  unsigned long a_time;
  struct super_block* a_sb;
  unsigned int a_acc;
} Acache;

#define C_FREE 0
#define C_USED 1
#define C_DIRTY_NORMAL 2
#define C_DIRTY_UNCOMPR 3
#define C_DELETED 4

typedef struct {
  unsigned int c_time;
  unsigned int c_flags;
  unsigned int c_count;
  unsigned int c_length;
  unsigned int c_clusternr;
  struct super_block* c_sb;
  unsigned char* c_data;
  struct wait_queue*c_wait;
  unsigned int c_locked;
} Cluster_head;


#ifdef __KERNEL__
void dbl_mdfat_value(struct super_block*sb, int clusternr,
                     Mdfat_entry*new,Mdfat_entry*mde);
int dbl_fat_nextcluster(struct super_block*sb,int clusternr,int*);
int dbl_bitfat_value(struct super_block*sb,int sektornr,int*);
void exit_dbl(struct super_block*sb);
int find_free_bitfat(struct super_block*sb, int sektornr, int size);
int replace_existing_cluster(struct super_block*sb, int cluster, 
                             int near_sector,
                             Mdfat_entry*);
#endif
int dbl_compress(unsigned char* clusterk, unsigned char* clusterd, 
                      int size, int method,int);
int stac_compress(void* pin,int lin, void* pout, int lout,
                  int method, int cfaktor);
#ifdef __KERNEL__
int dbl_decompress(unsigned char*clusterd, unsigned char*clusterk,
               Mdfat_entry*mde);
int dmsdos_write_cluster(struct super_block*sb,
                         unsigned char* clusterd, int length, int clusternr,
                         int near_sector, int ucflag);

#define CHS(i) ( (unsigned short)i[0]|(unsigned short)i[1]<<8 )
#define CHL(i) ( (unsigned long)i[0]|(unsigned long)i[1]<<8| \
                 (unsigned long)i[2]<<16|(unsigned long)i[3]<<24 )

int dbl_mdfat_cluster2sector(struct super_block*sb,int clusternr);
int simple_check(struct super_block*sb,int repair);
void do_spc_init(void);
void do_spc_exit(void);
void lock_mdfat_alloc(void);
void unlock_mdfat_alloc(void);
void free_cluster_sectors(struct super_block*sb, int clusternr);
void stac_special_free(struct super_block*sb, int clusternr);
int stac_write_cluster(struct super_block*sb,
                       unsigned char* clusterd, int length, int clusternr,
                       int near_sector, int ucflag);
int stac_read_cluster(struct super_block*sb,unsigned char*clusterd,
                      int clusternr);
void free_idle_cache(void);
void free_idle_ccache(void);
void ccache_init(void);
void free_ccache_dev(struct super_block*sb);
Cluster_head* ch_read(struct super_block*sb,int clusternr);
Cluster_head* ch_noread(struct super_block*sb,int clusternr);
Cluster_head* find_in_ccache(struct super_block*sb,
                             int clusternr,Cluster_head**lastfree,
                             Cluster_head**oldest);
int ch_dirty(Cluster_head*,int near,int ucflag);
void ch_dirty_retry_until_success(Cluster_head*,int near,int ucflag);
void ch_free(Cluster_head*);
void sync_cluster_cache(int allow_daemon);
void delete_cache_cluster(struct super_block*sb, int clusternr);
void log_list_statistics(void);
void log_ccache_statistics(void);
void log_found_statistics(void);
int sq_dec(void* pin,int lin, void* pout, int lout, int flg);
                                               
/* Stacker cluster allocation types access */

/* structure for walking/working with each sector of cluster */
typedef struct {
  struct super_block*sb;
  int clusternr;
  int start_sect;
  int start_len;
  int flags;
  int sect_cnt;
  int compressed;
  int bytes_in_last;
  int bytes_in_clust;
  struct buffer_head *fbh; /* first sector of fragmented clust */
  /* changes during fragments reads */
  int fcnt;	/* count of unreaded fragments */
  int flen;	/* rest sectors in fragment */
  int sect;	/* actual sector */
  int offset;	/* byte offset in sector */
  int bytes;	/* number of data bytes in sector */
  __u8* finfo;	/* points to actual field in fbh */
} Stac_cwalk;

int stac_cwalk_init(Stac_cwalk *cw,struct super_block*sb,
		 int clusternr,int flg);
int stac_cwalk_sector(Stac_cwalk *cw);
void stac_cwalk_done(Stac_cwalk *cw);

#ifdef __FOR_KERNEL_2_1_76
/* some hacks since the memcpy_from/tofs functions have changed here */
#define word short int
#define byte unsigned char
long get_fs_long(void* x);
word get_fs_word(void* x);
byte get_fs_byte(void* x);
void put_fs_long(long value, void* x);
void put_fs_word(word value, void* x);
void put_fs_byte(byte value, void* x);
#undef word
#undef byte
#define memcpy_fromfs copy_from_user
#define memcpy_tofs copy_to_user
#endif /* __FOR_KERNEL_2_1_76 */

/* loglevel defines */

extern unsigned long loglevel;

#ifdef SEQLOG
int log_prseq(void);
#define LOGCMD if(log_prseq())printk
#else
#define LOGCMD printk
#endif

#ifndef NOLOG
#define LOG_FS      if(loglevel&0x00000001)LOGCMD
#define LOG_CLUST   if(loglevel&0x00000002)LOGCMD
#define LOG_DIR     if(loglevel&0x00000004)LOGCMD
#define LOG_LLRW    if(loglevel&0x00000008)LOGCMD
#define LOG_DFAT    if(loglevel&0x00000010)LOGCMD
#define LOG_MDFAT   if(loglevel&0x00000020)LOGCMD
#define LOG_BITFAT  if(loglevel&0x00000040)LOGCMD
#define LOG_DECOMP  if(loglevel&0x00000080)LOGCMD
#define LOG_COMP    if(loglevel&0x00000100)LOGCMD
#define LOG_ALLOC   if(loglevel&0x00000200)LOGCMD
#define LOG_DAEMON  if(loglevel&0x00000400)LOGCMD
#define LOG_CCACHE  if(loglevel&0x00000800)LOGCMD
#define LOG_REST    if(loglevel&0x80000000)LOGCMD
#else
#define LOG_FS(x,args...)
#define LOG_CLUST(x,args...)
#define LOG_DIR(x,args...)
#define LOG_LLRW(x,args...)
#define LOG_DFAT(x,args...)
#define LOG_MDFAT(x,args...)
#define LOG_BITFAT(x,args...)
#define LOG_DECOMP(x,args...)
#define LOG_COMP(x,args...)
#define LOG_ALLOC(x,args...)
#define LOG_DAEMON(x,args...)
#define LOG_CCACHE(x,args...)
#define LOG_REST(x,args...)
#endif

#ifdef __FOR_KERNEL_2_1_76
/* some hacks since the memcpy_from/tofs functions have changed here */
#include  <asm/uaccess.h>
#define memcpy_fromfs copy_from_user
#define memcpy_tofs copy_to_user
#define get_fs_long(ptr) \
            ({ long  x; \
               __get_user_check(x,(long  *)(ptr),sizeof(long)); x; })
#define get_fs_word(ptr) \
            ({ short x; \
               __get_user_check(x,(short *)(ptr),sizeof(short)); x; })
#define get_fs_byte(ptr) \
            ({ unsigned char x; \
               __get_user_check(x,(unsigned char*)(ptr),sizeof(char)); x; })
#define put_fs_long(x,ptr) \
            ( __put_user_check((x),(long *)(ptr),sizeof(long)) )
#define put_fs_word(x,ptr) \
            ( __put_user_check((x),(short *)(ptr),sizeof(short)) )
#define put_fs_byte(x,ptr) \
            ( __put_user_check((x),(unsigned char*)(ptr),sizeof(char)) )
#endif /* __FOR_KERNEL_2_1_76 */

struct buffer_head *raw_bread (
        struct super_block *sb,
        int block);
struct buffer_head *raw_getblk (
        struct super_block *sb,
        int block);
void raw_brelse (
        struct super_block *sb,
        struct buffer_head *bh);
void raw_mark_buffer_dirty (
        struct super_block *sb,
        struct buffer_head *bh,
        int dirty_val);
void raw_set_uptodate (
        struct super_block *sb,
        struct buffer_head *bh,
        int val);
int raw_is_uptodate (
        struct super_block *sb,
        struct buffer_head *bh);
void raw_ll_rw_block (
        struct super_block *sb,
        int opr,
        int nbreq,
        struct buffer_head *bh[32]);

#define FAKED_ROOT_DIR_OFFSET 1
#define FAKED_DATA_START_OFFSET 1000
int dmsdos_read_cluster(struct super_block*sb,
                        unsigned char*clusterd, int clusternr);

struct buffer_head* dblspace_bread(struct super_block*sb,int vsector);
struct buffer_head *dblspace_getblk (
        struct super_block *sb,
        int block);
void dblspace_brelse(struct super_block* sb,struct buffer_head*bh);
void dblspace_mark_buffer_dirty(struct super_block*sb,struct buffer_head*bh,
                                int dirty_val);
void dblspace_set_uptodate (
        struct super_block *sb,
        struct buffer_head *bh,
        int val);
int dblspace_is_uptodate (
        struct super_block *sb,
        struct buffer_head *bh);
void dblspace_ll_rw_block (
        struct super_block *sb,
        int opr,
        int nbreq,
        struct buffer_head *bh[32]);
int stac_bitfat_state(struct super_block*sb,int new_state);
int dblspace_fat_access(struct super_block*sb, int clusternr,int newval);
int dblspace_bmap(struct inode*inode, int block);
int dblspace_smap(struct inode*inode, int block);
int ds_dec(void* pin,int lin, void* pout, int lout, int flg);
void dblspace_zero_new_cluster(struct inode*,int clusternr);
#ifdef __FOR_KERNEL_2_1_76
ssize_t dblspace_file_read(struct file *filp,char *buf,size_t count,
                           loff_t *ppos);
ssize_t dblspace_file_write(struct file *filp,const char *buf,size_t count,
                           loff_t *ppos);
int dblspace_mmap(struct file*file,
                  struct vm_area_struct*vma);
#else
int dblspace_file_read(struct inode *inode,struct file *filp,char *buf,
                                int count);
int dblspace_file_write(struct inode *inode,struct file *filp,const char *buf,
                                int count);
int dblspace_mmap(struct inode*inode,struct file*file,
                  struct vm_area_struct*vma);
#endif
#ifdef __FOR_KERNEL_2_1_78
int dblspace_readpage(struct dentry*dentry, struct page *page);
#else
int dblspace_readpage(struct inode *inode, struct page *page);
#endif
int dmsdos_ioctl_dir(struct inode *dir,struct file *filp,
                     unsigned int cmd, unsigned long data);
int try_daemon(struct super_block*sb,int clusternr, int length, int method);
void remove_from_daemon_list(struct super_block*sb,int clusternr);
void force_exit_daemon(void);
void dblspace_reada(struct super_block*sb, int sector,int count);
void init_reada_list(void);
void kill_reada_list_dev(int dev);

typedef struct
{ int clusternr;
  struct super_block*sb;
  int length;   /* in bytes */
  char flag;
  int method;
} Rwlist;

void init_daemon(void);
void exit_daemon(void);
void clear_list_dev(struct super_block*sb);

#endif /* __KERNEL__ */

#define DMSDOS_IOCTL_MIN 0x2000
#define DMSDOS_IOCTL_MAX 0x201F
#define DMSDOS_GET_DBLSB 0x2000
#define DMSDOS_EXTRA_STATFS 0x2001
#define DMSDOS_READ_BLOCK 0x2002
#define DMSDOS_WRITE_BLOCK 0x2003
#define DMSDOS_READ_DIRENTRY 0x2004
#define DMSDOS_WRITE_DIRENTRY 0x2005
#define DMSDOS_READ_BITFAT 0x2006
#define DMSDOS_WRITE_BITFAT 0x2007
#define DMSDOS_READ_MDFAT 0x2008
#define DMSDOS_WRITE_MDFAT 0x2009
#define DMSDOS_READ_DFAT 0x200a
#define DMSDOS_WRITE_DFAT 0x200b
#define DMSDOS_SET_COMP 0x200c
#define DMSDOS_SET_CF 0x200d
#define DMSDOS_SIMPLE_CHECK 0x200e
#define DMSDOS_DUMPCACHE 0x200f
#define DMSDOS_D_ASK 0x2010
#define DMSDOS_D_READ 0x2011
#define DMSDOS_D_WRITE 0x2012
#define DMSDOS_D_EXIT 0x2013
#define DMSDOS_MOVEBACK 0x2014
#define DMSDOS_SET_MAXCLUSTER 0x2015
#define DMSDOS_READ_CLUSTER 0x2016
#define DMSDOS_FREE_IDLE_CACHE 0x2017
#define DMSDOS_SET_LOGLEVEL 0x2018
#define DMSDOS_SYNC_CCACHE 0x2019
#define DMSDOS_LOG_STATISTICS 0x201a
#define DMSDOS_SET_SPEEDUP 0x201b
#define DMSDOS_RECOMPRESS 0x201c
#define IS_DMSDOS_IOCTL(cmd) ((cmd)>=DMSDOS_IOCTL_MIN&&(cmd)<=DMSDOS_IOCTL_MAX)

        
#endif
